const canvas =  document.getElementById('canvas_1');
const gl = canvas.getContext('webgl');

const glUtil = {
  shader:{
    loadCombined:function(vsSource, fsSource)
      {
        let vertexShader = glUtil.shader.loadSingle(gl.VERTEX_SHADER, vsSource);
        let fragmentShader = glUtil.shader.loadSingle(gl.FRAGMENT_SHADER, fsSource);
        let shaderComb = gl.createProgram();
        gl.attachShader(shaderComb, vertexShader);
        gl.attachShader(shaderComb, fragmentShader);
        gl.linkProgram(shaderComb);
        if (!gl.getProgramParameter(shaderComb, gl.LINK_STATUS)) {
          console.log("Shader could not be loaded: " + gl.getProgramInfoLog(shaderComb));
          return null;
        }
        return shaderComb;
      },
    loadSingle:function(type, source)
      {
        let shader = gl.createShader(type);
        gl.shaderSource(shader, source);
        gl.compileShader(shader);
        if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
          console.log("Shader could not be loaded: " + gl.getShaderInfoLog(shader));
          console.log(source);
          gl.deleteShader(shader);
          return null;
        }
        return shader;
      }
  },
  loadTexture:function(file)
    {
      let texture = gl.createTexture();
      gl.bindTexture(gl.TEXTURE_2D, texture);
      // Single Pixel stuff until image has loaded
      let level = 0;
      let internalFormat = gl.RGBA;
      let width = 1;
      let height = 1;
      let border = 0;
      let srcFormat = gl.RGBA;
      let srcType = gl.UNSIGNED_BYTE;
      gl.texImage2D(gl.TEXTURE_2D, level, internalFormat,
                width, height, border, srcFormat, srcType,
                new Uint8Array([0, 0, 0, 255]));
      // Actual image
      let image = new Image();
      image.onload = function() {
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texImage2D(gl.TEXTURE_2D, level, internalFormat,
                  srcFormat, srcType, image);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
        // Pixel graphics -> use nearest
        // Also means we don't care about mipmaps
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
      };
      image.crossOrigin = undefined;
      image.src = file;
      return texture;
    }
}
