var mineField = null;

var recentGenerator = null;

var timerValue = 0;

var validMineField = false;

var chordInfo = {
  lastId: -1,
  lastTime: 0,
  allowLeft: true,
  allowRight: true,
  timeOut: 400,
};

function generateNewGame(generator){
  allowDrawing = false;
  validMineField = false;
  recentGenerator = generator;
  // Deal with deleting shit
  deleteAllBuffers();
  // Adjust canvas size
  canvas.width = 32 * generator.sizeX;
  canvas.height = 32 * generator.sizeY;
  // Keep trying to generate a board if an attempt fails
  let fails = 0;
  while(!attemptGenerateNewGame(generator)){
    fails++;
    if(fails == setting_maxFails){
      alert("Setting could not be generated!");
      drawBlack();
      return;
    }
    console.log("Failed Generation Attempt");
  }
  // Create buffers
  createInitialBuffers();
  allowDrawing = true;
  validMineField = true;
  updateSideCounters();
  timerValue = 0;
  // Draw
  draw();
}

function attemptGenerateNewGame(generator){
  let patience = {
    a: setting_patience,
    lost: function() {this.a--; return this.a == 0;}
  };
  // Calcualate the size for the node grid
  const nodesPerGridCell = 5;
  let gridCellIdealEdge = Math.sqrt(nodesPerGridCell * generator.sizeX * generator.sizeY / generator.nodeCount);
  // Generate a new game
  mineField = {
    gameState: -1,
    properties: {
      sizeX: generator.sizeX,
      sizeY: generator.sizeY,
      nodeCount: generator.nodeCount,
      mines: generator.mines,
      gridSizeX: Math.ceil(generator.sizeX / gridCellIdealEdge), 
      gridSizeY: Math.ceil(generator.sizeY / gridCellIdealEdge), 
    },
    // List of all nodes
    nodes:[],
    // Place nodes in a 2D array 
    nodeGrid:[],
    // Get grid cell id
    getGridIdsOfPos: function(x, y) {
      return [
        clamp(0,
          Math.floor(this.properties.gridSizeX * x / this.properties.sizeX),
          this.properties.gridSizeX - 1),
        clamp(0,
          Math.floor(this.properties.gridSizeY * y / this.properties.sizeY),
          this.properties.gridSizeY - 1),
      ];
    },
    // Add a node
    pushNode: function(node) {
      this.nodes.push(node);
      let g = this.getGridIdsOfPos(node.posX, node.posY);
      this.nodeGrid[g[0]][g[1]].push(node);
    },
    // Get nodes in certain regions
    // Just get nodes in a certain region based on the grid
    getNodesRough: function(x, y, rad) {
      let n = [];
      let a = this.getGridIdsOfPos(x - rad, y - rad);
      let b = this.getGridIdsOfPos(x + rad, y + rad);
      for(let x = a[0]; x <= b[0]; x++){
        for(let y = a[1]; y <= b[1]; y++){
          n = n.concat(this.nodeGrid[x][y]);
        }
      }
      return n;
    },
    // Get the nodes within a cerain radius of a location
    getNodesRefined: function(x, y, radMin, radMax) {
      let n1 = this.getNodesRough(x, y, radMax);
      let n2 = [];
      radMax *= radMax;
      if(radMin > 0) radMin *= radMin;
      for(let n of n1){
        let dx = x - n.posX;
        let dy = y - n.posY;
        let r = dx * dx + dy * dy;
        if((r >= radMin) && (r <= radMax)) n2.push(n);
      }
      return n2;
    },
    // Was a node clicked and if so which one
    parseClickedNode: function(x, y) {
      let n = this.getNodesRefined(x / 32.0 - 0.5, y / 32.0 - 0.5, 0, 0.5);
      if(n.length == 0) return -1;
      return n[0].id;
    },
    // Can two node be connected
    connectable: function(a, b) {
      if(a == b) return;
      nodeA = this.nodes[a];
      nodeB = this.nodes[b];
      x0 = Math.min(nodeA.posX, nodeB.posX) - 0.75;
      x1 = Math.max(nodeA.posX, nodeB.posX) + 0.75;
      y0 = Math.min(nodeA.posY, nodeB.posY) - 0.75;
      y1 = Math.max(nodeA.posY, nodeB.posY) + 0.75;
      let gMin = this.getGridIdsOfPos(x0, y0);
      let gMax = this.getGridIdsOfPos(x1, y1);
      x0 = gMin[0];
      y0 = gMin[1];
      x1 = gMax[0];
      y1 = gMax[1];
      let v2_AB = [nodeB.posX - nodeA.posX, nodeB.posY - nodeA.posY];
      let v2_Tan = vecUtil.normalize(v2_AB);
      let v2_Nrm = [-v2_Tan[1], v2_Tan[0]];
      let len = vecUtil.dot(v2_Tan, v2_AB);
      for(let x = x0; x <= x1; x++){
        for(let y = y0; y <= y1; y++){
          let gNodes = this.nodeGrid[x][y];
          for(let nodeC of gNodes){
            if((nodeC.id != a) && (nodeC.id != b)){
              let v2_AC = [nodeC.posX - nodeA.posX, nodeC.posY - nodeA.posY];
              let t = vecUtil.dot(v2_AC, v2_Tan);
              let n = vecUtil.dot(v2_AC, v2_Nrm);
              if((t > 0) && (t < len) && (Math.abs(n) < 0.75)) return false;
            }
          }
        }
      }
      return true;
    },
    // Link two nodes
    linkNodes: function(a, b) {
      if(a == b) return;
      let na = this.nodes[a];
      let nb = this.nodes[b];
      for(n of na.connectedTo) if(n == b) return;
      na.connectedTo.push(b);
      nb.connectedTo.push(a);
    },
  };
  // Generate the node grid
  for(let i = 0; i < mineField.properties.gridSizeX; i++){
    mineField.nodeGrid.push([]);
    for(let j = 0; j < mineField.properties.gridSizeX; j++){
      mineField.nodeGrid[i].push([]);
    }
  }
  // Just generate some nodes at random positions for now
  let generated = 0;
  while(generated < mineField.properties.nodeCount){
    let genXY = generator.distrib(mineField.properties.sizeX, mineField.properties.sizeY);
    let genX = genXY[0];
    let genY = genXY[1];
    // Check if this position is too close to another node
    if(mineField.getNodesRefined(genX, genY, 0, 1.0).length == 0){
      mineField.pushNode(
        {
          id: generated,
          posX: genX,
          posY: genY,
          connectedTo:[],
          open:false,
          flag:false,
          mine:false,
          neighbours:0,
        }
      );
      generated++;
    }else{
      if(patience.lost()) return false;
    }
  }
  // This is a bad O(n^4) implementation (n being the number of nodes in one direction)
  // If you should need large n implement a better connection algorithm
  // Could probably be optimised to O(n^x) with x between 2 and 3
  // This will also be largely dependent on node density
  for(let i = 0; i < mineField.properties.nodeCount; i++){
    for(let j = i + 1; j < mineField.properties.nodeCount; j++){
      if(mineField.connectable(i, j)){
        mineField.linkNodes(i, j);
      }
    }
  }
  for(let node of mineField.nodes){
    node.connectable = node.connectedTo;
    node.connectedTo = [];
    nodeIdProximitySort(node.connectable, node.posX, node.posY);
  }
  let con = generator.connect;
  let con0 = Math.pow(con.rangeDecay, -con.allways + 1);
  for(let node of mineField.nodes){
    let a = con0;
    let i = 0;
    while((a > 0.01) && (i < node.connectable.length)){
      let exp = Math.min(1, con.preExponential
                * Math.exp(-con.distanceExponent
                  * nodeDistance(node, mineField.nodes[node.connectable[i]])));
      if(Math.random() < (a * exp)){
        mineField.linkNodes(node.id, node.connectable[i]);
      }
      a *= con.rangeDecay;
      i++;
    }
    i = 0;
    while((Math.random() < con.randDecay) && (i < node.connectable.length)){
      let t = Math.floor(Math.random() * node.connectable.length);
      let exp = Math.min(1, con.preExponential
                * Math.exp(-con.distanceExponent
                  * nodeDistance(node, mineField.nodes[node.connectable[t]])));
      if(exp > Math.random()) mineField.linkNodes(node.id, node.connectable[t]);
      i++;
    }
    node.connectable = null;
  }
  // Everything worked
  mineField.gameState = 0;
  return true;
}

function nodeDistance(a, b){
  let dx = a.posX - b.posX;
  let dy = a.posY - b.posY;
  return Math.sqrt(dx * dx + dy * dy);
}


function generateMines(id){
  let node0 = mineField.nodes[id];
  let safe = node0.connectedTo.concat(id);
  //console.log(safe);
  let minesPlaced = 0;
  while(minesPlaced < mineField.properties.mines){
    let i = Math.floor(Math.random() * mineField.properties.nodeCount);
    if(!mineField.nodes[i].mine){
      // Check if this tile should be safe
      let valid = true;
      for(let x of safe){
        if(x == i){
          valid = false;
          break;
        }
      }
      if(valid){
        mineField.nodes[i].mine = true;
        minesPlaced++;
      }
    }
  }
  for(let node of mineField.nodes){
    node.neighbours = 0;
    for(let i of node.connectedTo){
      if(mineField.nodes[i].mine) node.neighbours++;
    }
  }
  timerValue = (new Date()).getTime();
}

function dig(id){
  let node = mineField.nodes[id];
  if(node.open){
    if((mineField.gameState == 1) && chordInfo.allowLeft){
      if((id != chordInfo.lastId) || (((new Date()).getTime() - chordInfo.lastTime) > chordInfo.timeOut)){
        chordInfo.lastId = id;
        chordInfo.lastTime = (new Date()).getTime();
      }else{
        // Successfull double click
        // Check for flags
        let flags = 0;
        for(let x of node.connectedTo){
          let nx = mineField.nodes[x];
          if((!nx.open) && nx.flag) flags++;
        }
        if(flags == node.neighbours){
          for(let x of node.connectedTo){
            let nx = mineField.nodes[x];
            if((!nx.open) && (!nx.flag)) dig(x);
          }
        }
      }
    }
  }else{
    node.open = true;
    if(node.mine){
      mineField.gameState = 2;
      timerValue = (new Date()).getTime() - timerValue;
    }else{
      if(node.neighbours == 0){
        for(let x of node.connectedTo){
          if(!mineField.nodes[x].open){
            dig(x);
          }
        }
      }
    }
  }
}

// Sort array of nodes based on ther distance to a point
function nodeIdProximitySort(I, x, y){
  I.sort((i, j) =>
    {
      let ii = mineField.nodes[i];
      let jj = mineField.nodes[j];
      return ((Math.pow(ii.posX - x, 2) + Math.pow(ii.posY - y, 2)
              >= Math.pow(jj.posX - x, 2) + Math.pow(jj.posY - y, 2))
              ? 1 : -1)
    }
  );
}

// Update the counter for the flags and opened tiles
function printSingleCounter(a, c, t){
  a.div_tex.innerText = c + " / " + t;
}

function updateSideCounters(){
  let open = 0;
  let flags = 0;
  for(let node of mineField.nodes){
    if(node.open && (!node.mine)) open++;
    switch(mineField.gameState){
      case 0:
      case 1:
        if(node.flag && (!node.open)) flags++;
        break;
      case 2:
        if(node.flag && node.mine && (!node.open)) flags++; 
        break;
      case 3:
        if(!node.open) flags++;
        break;
    }
  }
  let mines = mineField.properties.mines;
  let safe = mineField.properties.nodeCount - mines;
  if(mineField.gameState == 1){
    if(safe == open){
      mineField.gameState = 3;
      flags = mines;
      timerValue = (new Date()).getTime() - timerValue;
    }
  }
  printSingleCounter(counterDisplay.floor, open, safe);
  printSingleCounter(counterDisplay.mine, flags, mines);
}

// Prevent the player from rightclicking the canvas
canvas.oncontextmenu = function (e) {
    e.preventDefault();
};

canvas.addEventListener('mouseup',
  function(event) {
    let i = mineField.parseClickedNode(
      event.pageX - canvas.getBoundingClientRect().left - 1 - window.pageXOffset,
      event.pageY - canvas.getBoundingClientRect().top - 1 - window.pageYOffset);
    if(document.selection && document.selection.empty) {
      document.selection.empty();
    } else if(window.getSelection) {
      let sel = window.getSelection();
      sel.removeAllRanges();
    }
    if(i == -1) return;
    if(isMobile) selectHighlight(i);
    //console.log("Clicked on " + i);
    let node = mineField.nodes[i];
    let lr = 0;
    if(isMobile){
      lr = mobileLeftclick ? 0 : 2;
    }else{
      lr = event.button;
    }
    switch(lr){
      case 0:
        // Leftclick
        if(!node.flag){
          switch(mineField.gameState){
            case 0:
              // Generate the mines
              generateMines(i);
              mineField.gameState = 1;
            case 1:
              // Dig
              dig(i);
              break;
          }
        }
        break;
      case 2:
        // Rightclick
        switch(mineField.gameState){
          case 0:
          case 1:
            if(node.open){
              if(chordInfo.allowRight){
                // Right click chording
                if((i != chordInfo.lastId) || (((new Date()).getTime() - chordInfo.lastTime) > chordInfo.timeOut)){
                  chordInfo.lastId = i;
                  chordInfo.lastTime = (new Date()).getTime();
                }else{
                  // Successfull double click
                  let closed = 0;
                  for(let x of node.connectedTo){
                    if(!mineField.nodes[x].open) closed++;
                  }
                  if(closed == node.neighbours){
                    for(let x of node.connectedTo){
                      if(!mineField.nodes[x].open) mineField.nodes[x].flag = true;
                    }
                  }
                }
              }
            }else{
              node.flag = !node.flag;
            }
            break;
        }
        break;
    }
    updateSideCounters();
    draw();
  }, false);

document.getElementById("restartButton").onclick = function(){
  selectHighlight(-1);
  if(recentGenerator != null) generateNewGame(recentGenerator);
};
