var buffers = {
  nodes: null,
  edges: null,
};

var allowDrawing = false;
var highlightedNode = -1;

function selectHighlight(h){
  if(h == highlightedNode) return;
  if(!allowDrawing) return;
  highlightedNode = h;
  draw();
}

function createInitialBuffers(){
  createInitialBuffers_Nodes();
  createInitialBuffers_Edges();
}

function createInitialBuffers_Nodes(){
  // ====== Nodes =======
  let indices = [];
  let positions = [];
  let texcoords = [];
  let nodeIds = [];
  let tsi = setting_textureSizeInv;
  for(let node of mineField.nodes){
    let j = node.id;
    let i = j * 4;
    indices.push(i, i + 1, i + 2, i, i + 2, i + 3);
    positions.push(
      node.posX, node.posY,
      node.posX + 1, node.posY,
      node.posX + 1, node.posY + 1,
      node.posX, node.posY + 1);
    texcoords.push(0, 0, tsi, 0, tsi, tsi, 0, tsi);
    nodeIds.push(j, j, j, j);
  }
  buffers.nodes = {
    indices: 0,
    positions: 0,
    texcoords: 0,
    nodeIds: 0,
    length: mineField.properties.nodeCount * 6
  };

  buffers.nodes.indices = gl.createBuffer();
  gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, buffers.nodes.indices);
  gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(indices), gl.STATIC_DRAW);
  
  buffers.nodes.positions = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.nodes.positions);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(positions), gl.STATIC_DRAW);
  buffers.nodes.texcoords = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.nodes.texcoords);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(texcoords), gl.STATIC_DRAW);
  buffers.nodes.nodeIds = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.nodes.nodeIds);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(nodeIds), gl.STATIC_DRAW);
}

function createInitialBuffers_Edges(){
  // ====== Edges ======
  let indices = [];
  let positions = [];
  let nodeIds = [];
  for(let i = 0; i < mineField.properties.nodeCount; i++){
    let n = mineField.nodes[i];
    positions.push(n.posX + 0.5, n.posY + 0.5);
    nodeIds.push(i, i);
  }
  let vertices = mineField.properties.nodeCount;
  let triangles = 0;
  for(let node of mineField.nodes){
    for(let con of node.connectedTo){
      if(con > node.id){
        let node2 = mineField.nodes[con];
        let center = [0.5 * (node.posX + node2.posX) + 0.5, 0.5 * (node.posY + node2.posY) + 0.5];
        let nrm = vecUtil.normalize([node2.posY - node.posY, node.posX - node2.posX]);
        let conWidth = 0.12;
        positions.push(center[0] - nrm[0] * conWidth, center[1] - nrm[1] * conWidth,
                      center[0] + nrm[0] * conWidth, center[1] + nrm[1] * conWidth);
        nodeIds.push(node.id, con, node.id, con);
        indices.push(node.id, vertices, vertices + 1, vertices + 1, vertices, con);
        triangles += 2;
        vertices += 2;
      }
    }
  }
  buffers.edges = {
    indices: 0,
    positions: 0,
    nodeIds: 0,
    length: triangles * 3,
  };
  
  buffers.edges.indices = gl.createBuffer();
  gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, buffers.edges.indices);
  gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(indices), gl.STATIC_DRAW);
  
  buffers.edges.positions = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.edges.positions);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(positions), gl.STATIC_DRAW);
  buffers.edges.nodeIds = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.edges.nodeIds);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(nodeIds), gl.STATIC_DRAW);
}

function deleteAllBuffers(){
  if(buffers.nodes == null) return;
  gl.deleteBuffer(buffers.nodes.indices);
  gl.deleteBuffer(buffers.nodes.positions);
  gl.deleteBuffer(buffers.nodes.texcoords);
  gl.deleteBuffer(buffers.nodes.nodeIds);
  buffers.nodes = null;
  gl.deleteBuffer(buffers.edges.indices);
  gl.deleteBuffer(buffers.edges.positions);
  gl.deleteBuffer(buffers.edges.nodeIds);
  buffers.edges = null;
}

var tileLocations = new Float32Array(setting_maxNodes * 2);

function updateUBO(){
  // Field Size
  gl.uniform2fv(shaders.nodeShader.uniformLocations.fieldSize,
                new Float32Array(
                  [mineField.properties.sizeX,
                  mineField.properties.sizeY]));
  // Tile Locations
  let i_Max = Math.min(setting_maxNodes, mineField.properties.nodeCount);
  for(let i = 0; i < i_Max; i++){
    let node = mineField.nodes[i];
    let pos = [0, 0];
    if(node.open){
      pos = ((node.mine) ? textureAtlas.mineOpen : textureAtlas.open(node.neighbours));
    }else{
      switch(mineField.gameState){
        case 0:
        case 1:
          pos = ((node.flag) ? textureAtlas.flag : textureAtlas.closed);
          break;
        case 2:
          pos = ((node.flag) ?
            ((node.mine) ? textureAtlas.mineFlagged : textureAtlas.flagWrong) :
            ((node.mine) ? textureAtlas.mineClosed : textureAtlas.closed));
          break;
        case 3:
          pos = textureAtlas.flagWin;
          break;
      }
    }
    tileLocations[i * 2] = setting_textureSizeInv * pos[0]; //Math.floor(setting_textureSize * Math.random());
    tileLocations[i * 2 + 1] = setting_textureSizeInv * pos[1]; //Math.floor(setting_textureSize * Math.random());
  }
  gl.uniform1fv(shaders.nodeShader.uniformLocations.tileLocations, tileLocations);
}

// Select the correct buffers in combination with the shader
function bufferAttributeSetup(buffer, shader){
  // Position
  gl.bindBuffer(gl.ARRAY_BUFFER, buffer.positions);
  gl.vertexAttribPointer(
      shader.attribLocations.position,
      2,        // Components
      gl.FLOAT, // Type
      false,    // Normalize
      0,        // Stride
      0);       // Offset
  gl.enableVertexAttribArray(
      shader.attribLocations.position);
  
  // Check if we are drawing nodes or edges
  if(!(buffer.texcoords === undefined)){
    // ===== Nodes =====
    // TexCoords
    gl.bindBuffer(gl.ARRAY_BUFFER, buffer.texcoords);
    gl.vertexAttribPointer(
        shader.attribLocations.texCoord,
        2,        // Components
        gl.FLOAT, // Type
        false,    // Normalize
        0,        // Stride
        0);       // Offset
    gl.enableVertexAttribArray(
        shader.attribLocations.texCoord);
    
    // NodeIds
    gl.bindBuffer(gl.ARRAY_BUFFER, buffer.nodeIds);
    gl.vertexAttribPointer(
        shader.attribLocations.nodeId,
        1,        // Components
        gl.FLOAT, // Type
        false,    // Normalize
        0,        // Stride
        0);       // Offset
    gl.enableVertexAttribArray(
        shader.attribLocations.nodeId);
  }else{
    // ===== Edges =====
    // NodeIds
    gl.bindBuffer(gl.ARRAY_BUFFER, buffer.nodeIds);
    gl.vertexAttribPointer(
        shader.attribLocations.nodeIds,
        2,        // Components
        gl.FLOAT, // Type
        false,    // Normalize
        0,        // Stride
        0);       // Offset
    gl.enableVertexAttribArray(
        shader.attribLocations.nodeIds);
  }

  // Indexed rendering
  gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, buffer.indices);

  // Select Shader
  gl.useProgram(shader.program);
    
}

function drawBlack(){
  gl.viewport(0, 0, gl.canvas.width, gl.canvas.height);
  gl.clearColor(0.0, 0.0, 0.0, 1.0);
  gl.clear(gl.COLOR_BUFFER_BIT);
}

function draw(){
  if(!allowDrawing) return;
  allowDrawing = false;
  
  gl.viewport(0, 0, gl.canvas.width, gl.canvas.height);
  // 628d8d (Background)
  gl.clearColor(
    0x62 / 255.0,
    0x8b / 255.0,
    0x8d / 255.0,
    1.0);
  // Gray
  //gl.clearColor(0.7, 0.7, 0.7, 1.0);
  gl.clear(gl.COLOR_BUFFER_BIT);
  gl.disable(gl.DEPTH_TEST);
  
  //
  bufferAttributeSetup(buffers.nodes, shaders.nodeShader);
  
  updateUBO();
  // Draw Connections
  bufferAttributeSetup(buffers.edges, shaders.edgeShader);
  
  gl.uniform2fv(shaders.edgeShader.uniformLocations.fieldSize,
                new Float32Array(
                  [mineField.properties.sizeX,
                  mineField.properties.sizeY]));
  gl.uniform1f(shaders.edgeShader.uniformLocations.highlight, highlightedNode);
  
  gl.drawElements(gl.TRIANGLES, buffers.edges.length, gl.UNSIGNED_SHORT, 0);
  
  // Setup Shader and Buffer
  bufferAttributeSetup(buffers.nodes, shaders.nodeShader);
  
  updateUBO();
  
  gl.activeTexture(gl.TEXTURE0);
  gl.bindTexture(gl.TEXTURE_2D, resources.tiles);
  gl.uniform1i(shaders.nodeShader.uniformLocations.texture, 0);
  
  gl.drawElements(gl.TRIANGLES, buffers.nodes.length, gl.UNSIGNED_SHORT, 0);
  
  allowDrawing = true;
}
